#!PERL_PATH
#
# Nuc atom table generator
#

use CGI qw/ -no-xhtml :standard/;
use English;

#
#  Global Variables
#
$input_nuc_seq = param('seq');
$star_format = param('star_file_format');
$example_value = param('example');
%isotope_list = ("C", 13,
                 "H", 1,
                 "N", 15,
                 "O", 17,
                 "P", 31);

$oops = 0;
$std_nuc = "ATCGUI";
$which_nuc = "";

#
#  generate_star_21_tables is used to output NMR-STAR 2.1
#  formated templates for the given atom sequence.
#
sub generate_star_21_tables 
{
  print "_Mol_residue_sequence\n;\n";
  for($i = 0; $i < $seq_len; $i += 20)
  {
    print substr($orig_seq,$i,20) . "\n";
  }
  print ";\n\n";

  print "loop_\n\t_Residue_seq_code\n\t_Residue_author_seq_code\n\t_Residue_label\n\n";
  
  for($i = 0; $i < $seq_len; $i++)
  {
    print $i + 1 . "   @   ";
    if(substr($orig_seq,$i,1) eq "@")
    {
      print "(unknown)";
    }
    else
    {
      print substr($orig_seq,$i,1);
    }
    if(($i + 1) % 5)
    {
      print "\t";
    }
    else
    {
      print "\n";
    }
  }
  
  print "\n\nstop_\n";
  
  if( param('ambiguity') eq "on")
  {
    print "\n###################################################################\n";
    print "#       Chemical Shift Ambiguity Index Value Definitions          #\n";
    print "#                                                                 #\n";
    print "#   Index Value            Definition                             #\n";
    print "#                                                                 #\n";
    print "#      1             Unique (geminal atoms and geminal methyl     #\n";
    print "#                         groups with identical chemical shifts   #\n";
    print "#                         are assumed to be assigned to           #\n";
    print "#                         stereospecific atoms)                   #\n";
    print "#      2             Ambiguity of geminal atoms or geminal methyl #\n";
    print "#                         proton groups (e.g. Asp HB2 and HB3     #\n";
    print "#                         protons, Leu CD1 and CD2 carbons, or    #\n";
    print "#                         Leu HD1 and HD2 methyl protons)         #\n";
    print "#      3             Aromatic atoms on opposite sides of          #\n";
    print "#                         symmetrical rings (e.g. Tyr HE1 and HE2 #\n";
    print "#                         protons)                                #\n";
    print "#      4             Intraresidue ambiguities (e.g. Lys HG and    #\n";
    print "#                         HD protons or Trp HZ2 and HZ3 protons)  #\n";
    print "#      5             Interresidue ambiguities (Lys 12 vs. Lys 27) #\n";
    print "#      6             Intermolecular ambiguities (e.g. Asp 31 CA   #\n";
    print "#                         in monomer 1 and Asp 31 CA in monomer 2 #\n";
    print "#                         of a asymmetrical homodimer, duplex DNA #\n";
    print "#                         assignments, or other assignments that  #\n";
    print "#                         may apply to atoms in one or more       #\n";
    print "#                         molecule in the molecular assembly)     #\n";
    print "#      9             Ambiguous, specific ambiguity not defined    #\n";
    print "#                                                                 #\n";
    print "###################################################################\n\n";
  }
  
  print "\n# INSTRUCTIONS\n";
  print "# 1) Replace the @ - signs with appropriate values.\n";
  print "# 2) Text comments concerning the assignments can be\n";
  print "#    supplied in the full deposition.\n";
  print "# 3) Feel free to add or delete rows to the table as needed.\n";
  print "#    The row numbers (_Atom_shift_assign_ID values) will be\n";
  print "#    re-assigned to sequential values by BMRB\n\n";
  print "\n# The atom table chosen for this sequence is:\n\n";
  print "loop_\n";
  print "     _Atom_shift_assign_ID\n";
  print "     _Residue_seq_code\n";
  print "     _Residue_label\n";
  print "     _Atom_name\n";
  print "     _Atom_type\n";
  print "     _Chem_shift_value\n";
  print "     _Chem_shift_value_error\n";
  print "     _Chem_shift_ambiguity_code\n\n";
  print "#\n#Atom\tResidue\n";
  print "#shift\tSeq\tResidue\tAtom\tAtom\tShift/\tError/\tAmbiguity\n";
  print "#assign\tcode\tType\tName\tType\tppm\tppm\tCode\n";
  print "#---------------------------------------------------------------\n#\n\n";
  
  $cnt = 1;
  for($i = 0; $i < $seq_len; $i++)
  {
    $i1 = $i + 1;
    $nuc = substr($seq,$i,1);
    if($nuc eq "@")
    {
      next;
    }
    if($nuc eq "X")
    {
      print "$cnt\t $i1\t@ \t@ \t@ \t@ \t@ \t@ #Please insert the atoms in this residue\n";
      $cnt++;
    }
    else
    {
      if($any_atom)
      {
        for($j = $starts{$nuc}; $j <= $ends{$nuc}; $j++)
        {
          print "$cnt\t $i1\t$nuc\t$atoms[$j] \t" . substr($atoms[$j],0,1) . "\t@ \t@ \t@ \n";
          $cnt++;
        }
      }
      else
      {
        for($j = $starts{$nuc}; $j <= $ends{$nuc}; $j++)
        {
          for($k = 0; $k < @which_atom; $k++)
          {
            if($atoms[$j] =~ /$which_atom[$k]/)
            {
              print "$cnt\t $i1\t$nuc\t$atoms[$j] \t" . substr($atoms[$j],0,1) . "\t@ \t@ \t@ \n";
              $cnt++;
              last;
            }
          }
        }
      }
    }
  }
  
  print <<EOEXAMPLE

stop_

# The following loop is used to define sets of Atom-shift assignment IDs that
# represent related ambiguous assignments taken from the above list of
# assigned chemical shifts.  Each element in the set should be separated by a
# comma, as shown in the example below, and is the assignment ID for a chemical
# shift assignment that has been given as ambiguity code of 4 or 5.  Each set
# indicates that the observed chemical shifts are related to the defined 
# atoms, but have not been assigned uniquely to a specific atom in the set.


loop_
  _Atom_shift_assign_ID_ambiguity

#
#    Sets of Atom-shift Assignment Ambiguities
               #              
#    ------------------------------------------
# Example:    5,4,7
#
                @
stop_

EOEXAMPLE
} # end of generate_star_21_tables subroutine.

#
#  generate_star_31_tables is used to output NMR-STAR 3.1
#  formated templates for the given nuc sequence.
#
sub generate_star_31_tables 
{
  print "_Entity.Polymer_seq_one_letter_code\n;\n";
  for($i = 0; $i < $seq_len; $i += 20)
  {
    print substr($orig_seq,$i,20) . "\n";
  }
  print ";\n\n";

  print "loop_\n\t_Entity_comp_index.ID\n\t_Entity_comp_index.Auth_seq_ID\n\t_Entity_comp_index.Comp_ID\n\n";
  
  for($i = 0; $i < $seq_len; $i++)
  {
    print $i + 1 . "\t@ \t";
    if(substr($orig_seq,$i,1) eq "@")
    {
      print "(unknown)";
    }
    else
    {
      print substr($orig_seq,$i,1);
    }
#    print "\tNEED_ACC_NUM\t1\n";
    print "\n";

  }
  
  print "\n\nstop_\n";
  
  if( param('ambiguity') eq "on")
  {
    print "\n###################################################################\n";
    print "#       Chemical Shift Ambiguity Index Value Definitions          #\n";
    print "#                                                                 #\n";
    print "#   Index Value            Definition                             #\n";
    print "#                                                                 #\n";
    print "#      1             Unique (geminal atoms and geminal methyl     #\n";
    print "#                         groups with identical chemical shifts   #\n";
    print "#                         are assumed to be assigned to           #\n";
    print "#                         stereospecific atoms)                   #\n";
    print "#      2             Ambiguity of geminal atoms or geminal methyl #\n";
    print "#                         proton groups (e.g. Asp HB2 and HB3     #\n";
    print "#                         protons, Leu CD1 and CD2 carbons, or    #\n";
    print "#                         Leu HD1 and HD2 methyl protons)         #\n";
    print "#      3             Aromatic atoms on opposite sides of          #\n";
    print "#                         symmetrical rings (e.g. Tyr HE1 and HE2 #\n";
    print "#                         protons)                                #\n";
    print "#      4             Intraresidue ambiguities (e.g. Lys HG and    #\n";
    print "#                         HD protons or Trp HZ2 and HZ3 protons)  #\n";
    print "#      5             Interresidue ambiguities (Lys 12 vs. Lys 27) #\n";
    print "#      6             Intermolecular ambiguities (e.g. Asp 31 CA   #\n";
    print "#                         in monomer 1 and Asp 31 CA in monomer 2 #\n";
    print "#                         of a asymmetrical homodimer, duplex DNA #\n";
    print "#                         assignments, or other assignments that  #\n";
    print "#                         may apply to atoms in one or more       #\n";
    print "#                         molecule in the molecular assembly)     #\n";
    print "#      9             Ambiguous, specific ambiguity not defined    #\n";
    print "#                                                                 #\n";
    print "###################################################################\n\n";
  }
  
  print "\n# INSTRUCTIONS\n";
  print "# 1) Replace the @-signs with appropriate values.\n";
  print "# 2) Text comments concerning the assignments can be\n";
  print "#    supplied in the full deposition.\n";
  print "# 3) Feel free to add or delete rows to the table as needed.\n";
  print "#    The row numbers (_Atom_chem_shift.ID values)\n";
  print "#    will be re-assigned to sequential values by BMRB\n#\n";
#  print "#    The last two columns are for use by BMRB annotators\n";
#  print "#    and should be left unchanged.\n\n";
  print "\n# The atom table chosen for this sequence is:\n\n";
  print "loop_\n";
  print "     _Atom_chem_shift.ID\n";
  print "     _Atom_chem_shift.Comp_index_ID\n";
  print "     _Atom_chem_shift.Comp_ID\n";
  print "     _Atom_chem_shift.Atom_ID\n";
  print "     _Atom_chem_shift.Atom_type\n";
  print "     _Atom_chem_shift.Val\n";
  print "     _Atom_chem_shift.Val_err\n";
  print "     _Atom_chem_shift.Ambiguity_code\n";
  print "     _Atom_chem_shift.Occupancy\n";
  print"\n";
  
  
  print "#\n#\tComp\t\t\t\tChem\tC.S.\tC.S.\tChem\t\n";
  print "#\tindex\tComp\tAtom\tAtom\tshift\tval\tambig\tshift\t\n";
  print "#ID\tID\tID\tID\ttype\tval\terr\tcode\toccup\t\n";
  print "#-----------------------------------------------\n#\n\n";
  
  $cnt = 1;
  for($i = 0; $i < $seq_len; $i++)
  {
    $i1 = $i + 1;
    $nuc = substr($seq,$i,1);
    if($nuc eq "@")
    {
      next;
    }
    if($nuc eq "X")
    {
#      print " $cnt\t1\t1\t $i1\t $i1\t@ \t@ \t@ \t@ \t@ \t@ \t@ \t@ \t $i1 \t ?\t? \t?@\tNEED_ACC_NUM\t1  #Please insert the atoms in this residue\n";
      print " $cnt\t $i1\t@ \t@ \t@ \t@ \t@ \t@ \t ?  #Please insert the atoms in this residue\n";
      $cnt++;
    }
    else
    {
      if($any_atom)
      {
        for($j = $starts{$nuc}; $j <= $ends{$nuc}; $j++)
        {
          $atom_type = substr($atoms[$j],0,1);
          if ($atom_type eq "\"")
          {
            $atom_type = substr($atoms[$j],1,1);
          }
          if ($atom_type eq "\"")
          {
            $atom_type = substr($atoms[$j],1,1);
          }
          # $isotope = $isotope_list{ $atom_type };

#          print " $cnt\t1\t1\t $i1\t $i1\t$nuc\t$atoms[$j] \t" . $atom_type . "\t $isotope\t@ \t@ \t@ \t@ \t@ \t $i1\t? \t?\tNEED_ACC_NUM\t1\n";
          print " $cnt\t $i1\t$nuc\t$atoms[$j] \t" . $atom_type . "\t@ \t@ \t@ \t@ \t \n";
          $cnt++;
        }
      }
      else
      {
        for($j = $starts{$nuc}; $j <= $ends{$nuc}; $j++)
        {
          for($k = 0; $k < @which_atom; $k++)
          {
            $atom_type = substr($atoms[$j],0,1);
            if ($atom_type eq "\"")
            {
              $atom_type = substr($atoms[$j],1,1);
            }
            if ($atom_type eq "\"")
            {
              $atom_type = substr($atoms[$j],1,1);
            }
	    $isotope = $isotope_list{ $atom_type };

            if($atoms[$j] =~ /$which_atom[$k]/)
            {
              # print " $cnt\t1\t1\t $i1\t $i1\t$nuc\t$atoms[$j] \t" . $atom_type . "\t $isotope\t@ \t@ \t@ \t@ \t@ \t $i1\t? \t? \tNEED_ACC_NUM\t1\n";
              print " $cnt\t $i1\t$nuc\t$atoms[$j] \t" . $atom_type . "\t@ \t@ \t@ \t@ \t \n";
              $cnt++;
              last;
            }
          }
        }
      }
    }
  }
  
  print <<EOEXAMPLE

stop_

# The following table is used to define sets of ambiguous Atom-shift 
# assignments with ambiguity codes of 4 or 5 taken from the above list 
# of assigned chemical shifts.  Each set of ambiguous chemical shift 
# assignments should be given a unique 'Set_ID'. As shown in the example 
# below, the 'Set_ID' is given in the first column of the table and the 
# 'Atom_chem_shift.ID' for the chemical shift assignment that is a 
# member of the set is given in the second column (should have an 
# ambiguity code of 4 or 5).  In the example, the first set (Set_ID = 1) 
# has two members with Atom_chem_shift.IDs 158 and 171).  Each set 
# indicates that the observed chemical shifts are related to the defined 
# atoms, but have not been assigned uniquely to a specific atom in the 
# set.

loop_
   _Ambiguous_atom_chem_shift.Ambiguous_shift_set_ID
   _Ambiguous_atom_chem_shift.Atom_chem_shift_ID


#
#            Set_ID              Atom_chem_shift_ID

                @                    @
                @                    @

#
#    ------------------------------------------
# Example:    1                    158
#             1                    171
#             2                    234
#             2                    257

stop_

EOEXAMPLE

} # end of generate_star_31_tables subroutine.

#####################################################
#
#                       MAIN 
#
#####################################################

if ( ! ($example_value eq "") ) 
{
  $input_nuc_seq = "ATGA";
  $star_format = $example_value;
}
if(!($input_nuc_seq =~ /.+/))
{
  $oops = 1;
  $err = "Please enter a sequence of bases.";
}

for($i = 0; $i < length( $std_nuc ); $i++)
{
  if(param(substr($std_nuc,$i,1)) eq "on")
  {
    $which_nuc = $which_nuc . substr($std_nuc,$i,1);
  }
}

#convert sequence to uppercase and remove anything that isn't a std nuc
#code

$_ = $input_nuc_seq;
tr/a-z/A-Z/;
s/[\x00-\x20]//g;
$unchecked_seq = $_;
s/[^$std_nuc]/@/g;
$orig_seq = $_;

$bad_nuc = tr/@//;
if($bad_nuc && !$oops)
{
  $oops = 1;
  $idx = index($_, "@");
  $idx1 = $idx +1;
  $err = "The input sequence contains an unrecognized residue code \"" . 
    substr($unchecked_seq, $idx, 1) . "\" at position $idx1 of the sequence:<br> " .
    substr($unchecked_seq, 0, $idx) . "<b>" . substr($unchecked_seq, $idx, 1) . "</b>...";
}

# if the user wants only some nuc's, remove the list from the sequence.

if($which_nuc)
{
  if(param('invert_res') eq "on")
  {
    s/[$which_nuc]/@/g;
  }
  else
  {
    s/[^$which_nuc]/@/g;
  }
}

$seq_len = tr/A-Z@//;
$good_len = tr/A-Z//;
$seq = $_;

if(!$good_len && !$oops) 
{
  $oops = 1;
  $err = "The output has been limited to specific residues, but none of those" .
    " residues are in the input sequence.";
}

if($oops)
{
  print "Content-type: text/html\n\n";
  print "<html><head><title>Error</title></head><body>";
  print "<h3>Atom List Generator Error</h3>\n";
  print $err;
  print "</html></body>";
  exit;
}
@which_atom = ();
$any_atom = 1;

#
# build list of allowed atoms
#
if( param('call') eq "on")
{
  push(@which_atom, "C.*");
  $any_atom = 0;
}
if( param('cs') eq "on")
{
  push(@which_atom, "C.*'");
  $any_atom = 0;
}
if( param('cb') eq "on")
{
  push(@which_atom, "C[^']*\$");
  $any_atom = 0;
}
if( param('nall') eq "on")
{
  push(@which_atom, "N.*");
  $any_atom = 0;
}
if( param('hall') eq "on")
{
  push(@which_atom, "H.*");
  $any_atom = 0;
}
if( param('pall') eq "on")
{
  push(@which_atom, "^P.*");
  $any_atom = 0;
}
if( param('heavy') eq "on")
{
  push(@which_atom, "^[CNP].*");
  $any_atom = 0;
}

#if( param('all') eq "on")
#{
#  $any_atom = 1;
#}

$last_nuc = "";
$i = 0;
@atoms = ();
@chiralities = ();

#
# read amino acid atom file
#
open(AA_FILE, "BMRB_ADMIN_PATH/atom_gen/nuc_atom");
while($line = <AA_FILE>)
{
  chop($line);
  if(substr($line,0,1) eq "#")  # then the line is a comment
  {
    next;
  }

  @fields = split(/\t/,$line);
  if(($fields[3] ne "1") && ( param('all') eq "nmr"))  #then the line doesn't contain an atom for shift assignments
  {
    next;
  }
  if(($fields[4] ne "1") && ( param('all') eq "all")) #then the line doesn't contain a structural atom
  {
    next;
  }
  if(($fields[5] ne "1") && ( param('nuc_type') eq "DNA")) #then the line doesn't contain a DNA atom
  {
    next;
  }
  if(($fields[6] ne "1") && ( param('nuc_type') eq "RNA")) #then the line doesn't contain a RNA atom
  {
    next;
  }

  $atoms[$i] = $fields[1];
  $chiralities[$i] = $fields[2];
  if($fields[0] ne $last_nuc)
  {
    if($last_nuc)
    {
      $ends{$last_nuc} = $i - 1;
    }
    $starts{$fields[0]} = $i;
    $last_nuc = $fields[0];
  }
  $i++;
}

#
# read amino acid 3-letter code file
#
#open(AA_NAME, "BMRB_ADMIN_PATH/atom_gen/nuc_abbrv");
#while($line = <AA_NAME>)
#{
#  chop($line);
#  @fields = split(/\t/, $line);
#  $nuc_codes{$fields[0]} = $fields[1];
#}

$ends{$last_nuc} = $i - 1;

#
# write atoms to stdout
#

if ( ! ($example_value eq "") ) 
{
  $win_title = "NMR-STAR " . $example_value . " format example";
  print header,
      start_html( -lang=>'en_US', -title=>$win_title ),
      "\n";
  print "<HR size \"4\"><H1 align=\"center\">Examples of output from the Table Generator</H1>\n";
  print "<p><b><h2 align=\"center\">NMR-STAR $example_value format</h2></b></p><P align=\"center\">\n";
  print "the initial sequence was: <STRONG>$input_nuc_seq</STRONG>\n";
  print "<hr size \"4\">\n";
  print "<PRE>\n";
}
else
{
  print "Content-type: text/plain\n\n";
}

print "#\n# The original sequence entered was:\n";
print "# $orig_seq\n#\n";

print "\n\n# Expressed in NMR-STAR " . $star_format . ", this sequence is:\n\n";

if ( $star_format eq "2.1" )
{
  &generate_star_21_tables;
}
elsif ( $star_format eq "3.1" )
{
  &generate_star_31_tables;
}

if ( $example_value eq "")
{
  exit;
}
else {
  print "<\PRE>\n";
  print end_html,
    "\n";
}
